"""This module contains the main loop for updating sprites and running their events."""

import math as _math

from .mouse_loop import mouse_state
from ..callback import callback_manager, CallbackType
from ..callback.callback_helpers import run_callback, run_async_callback
from ..globals import globals_list
from ..io.screen import convert_pos, PYGAME_DISPLAY
from ..io.mouse import mouse
from ..objects.line import Line
from ..objects.sprite import point_touching_sprite


async def _update_sprites(do_events: bool = True):  # pylint: disable=too-many-branches
    # pylint: disable=too-many-nested-blocks
    globals_list.sprites_group.update()

    for sprite in globals_list.sprites_group.sprites():
        ######################################################
        # update sprites with results of physics simulation
        ######################################################
        if sprite.physics and sprite.physics.can_move:
            body = sprite.physics._pymunk_body
            angle = _math.degrees(body.angle)
            if isinstance(sprite, Line):
                sprite._x = body.position.x - (sprite.length / 2) * _math.cos(angle)
                sprite._y = body.position.y - (sprite.length / 2) * _math.sin(angle)
                sprite._x1 = body.position.x + (sprite.length / 2) * _math.cos(angle)
                sprite._y1 = body.position.y + (sprite.length / 2) * _math.sin(angle)
                # sprite._length, sprite._angle = sprite._calc_length_angle()
            else:
                if (
                    str(body.position.x) != "nan"
                ):  # this condition can happen when changing sprite.physics.can_move
                    sprite._x = body.position.x
                if str(body.position.y) != "nan":
                    sprite._y = body.position.y

            sprite.angle = (
                angle  # needs to be .angle, not ._angle so surface gets recalculated
            )
            sprite.physics._x_speed, sprite.physics._y_speed = body.velocity

        sprite._is_clicked = False
        if sprite.is_hidden:
            continue

        if not do_events and not sprite.physics:
            continue

        #################################
        # All @sprite.when_touching events
        #################################
        if sprite._touching_callback[0]:
            await run_async_callback(sprite._touching_callback[0], [], [])
        if sprite._touching_callback[1]:
            await run_async_callback(sprite._touching_callback[1], [], [])

        #################################
        # @sprite.when_clicked events
        #################################
        if mouse.is_clicked:
            if (
                point_touching_sprite(convert_pos(mouse.x, mouse.y), sprite)
                and mouse_state.click_happened_this_frame
            ):
                # only run sprite clicks on the frame the mouse was clicked
                sprite._is_clicked = True
                if callback_manager.get_callback(
                    CallbackType.WHEN_CLICKED_SPRITE, id(sprite)
                ):
                    for callback in callback_manager.get_callback(
                        CallbackType.WHEN_CLICKED_SPRITE, id(sprite)
                    ):
                        if not callback.is_running:
                            run_callback(
                                callback,
                                [],
                                [],
                            )

    globals_list.sprites_group.update()
    globals_list.sprites_group.draw(PYGAME_DISPLAY)



"""This module contains the function that simulates the physics of the game"""

from .sprites_loop import _update_sprites
from ..globals import globals_list
from ..physics import physics_space, _NUM_SIMULATION_STEPS


async def simulate_physics():
    """
    Simulate the physics of the game
    """
    # more steps means more accurate simulation, but more processing time
    for _ in range(_NUM_SIMULATION_STEPS):
        physics_space.step(1 / (globals_list.FRAME_RATE * _NUM_SIMULATION_STEPS))
        if not _ == _NUM_SIMULATION_STEPS - 1:
            await _update_sprites(False)



"""
This module contains the API for the game.
"""

from pygame import init  # pylint: disable=no-name-in-module,import-error

from .generators import new_text, new_box, new_circle, new_line, new_image, new_sound
from .events import (
    when_program_starts,
    repeat_forever,
    when_sprite_clicked,
    when_any_key_pressed,
    when_key_pressed,
    when_any_key_released,
    when_key_released,
    when_mouse_clicked,
    when_click_released,
)
from .utils import (
    start_program,
    stop_program,
    animate,
    set_backdrop,
    set_backdrop_image,
    timer,
    key_is_pressed,
    set_physics_simulation_steps,
)
from .random import random_number, random_color, random_position

init()



"""Collision callbacks for sprites."""

try:
    from enum import EnumType
except ImportError:
    from enum import (
        EnumMeta as EnumType,
    )  # In Python 3.10 the alias for EnumMeta doesn't yet exist
from pymunk import CollisionHandler

from play.callback.callback_helpers import run_callback
from play.physics import physics_space


class CollisionType(EnumType):
    SPRITE = 0
    WALL = 1


class CollisionCallbackRegistry:  # pylint: disable=too-few-public-methods
    """
    A registry for collision callbacks.
    """

    def _handle_collision(self, arbiter, _, __):
        shape_a, shape_b = arbiter.shapes
        if shape_a.collision_type in self.callbacks[True]:
            for shape_b_id, callback in self.callbacks[True][
                shape_a.collision_type
            ].items():
                if shape_b_id == shape_b.collision_type:
                    self.shape_registry[shape_a.collision_type]._touching_callback[
                        shape_a.actual_collision_type
                    ] = callback

        if shape_b.collision_type in self.callbacks[True]:
            for shape_a_id, callback in self.callbacks[True][
                shape_b.collision_type
            ].items():
                if shape_a_id == shape_a.collision_type:
                    self.shape_registry[shape_b.collision_type]._touching_callback[
                        shape_b.actual_collision_type
                    ] = callback
        return True

    def _handle_end_collision(self, arbiter, _, __):
        shape_a, shape_b = arbiter.shapes

        if shape_a.collision_type in self.callbacks[True]:
            for shape_b_id, callback in self.callbacks[True][
                shape_a.collision_type
            ].items():
                if shape_b_id == shape_b.collision_type:
                    self.shape_registry[shape_a.collision_type]._touching_callback[
                        shape_a.actual_collision_type
                    ] = None

        if shape_b.collision_type in self.callbacks[True]:
            for shape_a_id, callback in self.callbacks[True][
                shape_b.collision_type
            ].items():
                if shape_a_id == shape_a.collision_type:
                    self.shape_registry[shape_b.collision_type]._touching_callback[
                        shape_b.actual_collision_type
                    ] = None

        if shape_a.collision_type in self.callbacks[False]:
            for shape_b_id, callback in self.callbacks[False][
                shape_a.collision_type
            ].items():
                if shape_b_id == shape_b.collision_type:
                    run_callback(callback, [], [])
        if shape_b.collision_type in self.callbacks[False]:
            for shape_a_id, callback in self.callbacks[False][
                shape_b.collision_type
            ].items():
                if shape_a_id == shape_a.collision_type:
                    run_callback(callback, [], [])

        return True

    def __init__(self):
        self.callbacks = {True: {}, False: {}}
        self.shape_registry = {}
        handler: CollisionHandler = physics_space.add_default_collision_handler()
        handler.begin = self._handle_collision
        handler.separate = self._handle_end_collision

    def register(
        self, sprite_a, shape_a, shape_b, callback, collision_type, begin=True
    ):
        """
        Register a callback with a name.
        """
        shape_a.collision_type = id(shape_a)
        shape_b.collision_type = id(shape_b)
        self.shape_registry[shape_a.collision_type] = sprite_a
        shape_a.actual_collision_type = collision_type

        if not shape_a.collision_type in self.callbacks[begin]:
            self.callbacks[begin][shape_a.collision_type] = {}
        if shape_b.collision_type in self.callbacks[begin][shape_a.collision_type]:
            raise ValueError(f"Callback already registered for {shape_a} and {shape_b}")
        self.callbacks[begin][shape_a.collision_type][shape_b.collision_type] = callback


collision_registry = CollisionCallbackRegistry()



"""Core game loop and event handling functions."""

import pygame  # pylint: disable=import-error

from .controller_loop import (
    controller_axis_moved,
    controller_button_pressed,
    controller_button_released,
    _handle_controller,
    _handle_controller_events,
)
from .game_loop_wrapper import listen_to_failure
from .mouse_loop import _handle_mouse_loop, mouse_state
from .physics_loop import simulate_physics
from .sprites_loop import _update_sprites
from ..callback import callback_manager, CallbackType
from ..callback.callback_helpers import run_async_callback, run_callback
from ..globals import globals_list
from ..io.screen import screen, PYGAME_DISPLAY
from ..io.keypress import (
    key_num_to_name as _pygame_key_to_name,
    _keys_released_this_frame,
    _keys_to_skip,
    _pressed_keys,
)  # don't pollute user-facing namespace with library internals
from ..io.mouse import mouse
from ..loop import loop as _loop

_clock = pygame.time.Clock()


def _handle_pygame_events():
    """Handle pygame events in the game loop."""
    for event in pygame.event.get():
        if event.type == pygame.QUIT or (  # pylint: disable=no-member
            event.type == pygame.KEYDOWN  # pylint: disable=no-member
            and event.key == pygame.K_q  # pylint: disable=no-member
            and (
                pygame.key.get_mods() & pygame.KMOD_META  # pylint: disable=no-member
                or pygame.key.get_mods() & pygame.KMOD_CTRL  # pylint: disable=no-member
            )
        ):
            # quitting by clicking window's close button or pressing ctrl+q / command+q
            _loop.stop()
            return False
        if event.type == pygame.MOUSEBUTTONDOWN:  # pylint: disable=no-member
            mouse_state.click_happened_this_frame = True
            mouse._is_clicked = True
        if event.type == pygame.MOUSEBUTTONUP:  # pylint: disable=no-member
            mouse_state.click_release_happened_this_frame = True
            mouse._is_clicked = False
        if event.type == pygame.MOUSEMOTION:  # pylint: disable=no-member
            mouse.x, mouse.y = (event.pos[0] - screen.width / 2.0), (
                screen.height / 2.0 - event.pos[1]
            )
        _handle_controller_events(event)
        if event.type == pygame.KEYDOWN:  # pylint: disable=no-member
            if event.key not in _keys_to_skip:
                name = _pygame_key_to_name(event)
                if name not in _pressed_keys:
                    _pressed_keys.append(name)
        if event.type == pygame.KEYUP:  # pylint: disable=no-member
            name = _pygame_key_to_name(event)
            if not (event.key in _keys_to_skip) and name in _pressed_keys:
                _keys_released_this_frame.append(name)
                _pressed_keys.remove(name)
    return True


# pylint: disable=too-many-branches
async def _handle_keyboard():
    """Handle keyboard events in the game loop."""
    ############################################################
    # @when_any_key_pressed and @when_key_pressed callbacks
    ############################################################
    if (
        _pressed_keys
        and callback_manager.get_callbacks(CallbackType.PRESSED_KEYS) is not None
    ):
        press_subscription = callback_manager.get_callbacks(CallbackType.PRESSED_KEYS)
        for key in _pressed_keys:
            if key in callback_manager.get_callbacks(CallbackType.PRESSED_KEYS):
                for callback in press_subscription[key]:
                    if not callback.is_running:
                        await run_async_callback(
                            callback,
                            ["key"],
                            [],
                            key,
                        )
            if "any" in press_subscription:
                for callback in press_subscription["any"]:
                    if not callback.is_running:
                        await run_async_callback(
                            callback,
                            ["key"],
                            [],
                            key,
                        )
        keys_hash = hash(frozenset(_pressed_keys))
        if keys_hash in press_subscription:
            for callback in press_subscription[keys_hash]:
                if not callback.is_running:
                    await run_async_callback(
                        callback,
                        ["key"],
                        [],
                        _pressed_keys,
                    )

    ############################################################
    # @when_any_key_released and @when_key_released callbacks
    ############################################################
    if _keys_released_this_frame and callback_manager.get_callbacks(
        CallbackType.RELEASED_KEYS
    ):
        release_subscriptions = callback_manager.get_callbacks(
            CallbackType.RELEASED_KEYS
        )
        for key in _keys_released_this_frame:
            if key in release_subscriptions:
                for callback in release_subscriptions[key]:
                    if not callback.is_running:
                        await run_async_callback(
                            callback,
                            ["key"],
                            [],
                            key,
                        )
            if "any" in release_subscriptions:
                for callback in release_subscriptions["any"]:
                    if not callback.is_running:
                        await run_async_callback(
                            callback,
                            ["key"],
                            [],
                            key,
                        )


# pylint: disable=too-many-branches, too-many-statements
@listen_to_failure()
async def game_loop():
    """The main game loop."""
    _keys_released_this_frame.clear()
    mouse_state.click_happened_this_frame = False
    mouse_state.click_release_happened_this_frame = False

    _clock.tick(globals_list.FRAME_RATE)

    if not _handle_pygame_events():
        return

    await _handle_keyboard()

    if (
        mouse_state.click_happened_this_frame
        or mouse_state.click_release_happened_this_frame
    ):
        await _handle_mouse_loop()

    await _handle_controller()

    #############################
    # @repeat_forever callbacks
    #############################
    if callback_manager.get_callbacks(CallbackType.REPEAT_FOREVER) is not None:
        for callback in callback_manager.get_callbacks(CallbackType.REPEAT_FOREVER):
            if not callback.is_running:
                run_callback(callback, [], [])

    #############################
    # physics simulation
    #############################
    await simulate_physics()

    if globals_list.backdrop_type == "color":
        PYGAME_DISPLAY.fill(globals_list.backdrop)
    elif globals_list.backdrop_type == "image":
        PYGAME_DISPLAY.blit(
            globals_list.backdrop,
            (0, 0),
        )
    else:
        PYGAME_DISPLAY.fill((255, 255, 255))

    await _update_sprites()

    pygame.display.flip()
    _loop.create_task(game_loop())  # Call self again



"""Global variables for the game"""

import pygame


class Globals:  # pylint: disable=too-few-public-methods
    all_sprites = []
    sprites_group = pygame.sprite.Group()

    walls = []

    backdrop_type = "color"  # color or image
    backdrop = (255, 255, 255)

    FRAME_RATE = 60
    WIDTH = 800
    HEIGHT = 600


globals_list = Globals()



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

with open(HERE / 'requirements.txt') as f:
    required = f.read().splitlines()

# This call to setup() does all the work
setup(
    name="corderius-play",
    version="2.6.1",
    description="The easiest way to make games and media projects in Python.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Corderius-College-Amersfoort/play",
    author="koen1711",
    author_email="koenvurk1711@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    python_requires=">=3.10",
    packages=find_packages(),
    include_package_data=True,
    install_requires=required,
)



"""This module contains the mouse loop."""

from ..callback import callback_manager, CallbackType
from ..callback.callback_helpers import run_async_callback


class MouseState:  # pylint: disable=too-few-public-methods
    click_happened_this_frame = False  # pylint: disable=invalid-name
    click_release_happened_this_frame = False  # pylint: disable=invalid-name


mouse_state = MouseState()


async def _handle_mouse_loop():
    """Handle mouse events in the game loop."""
    ####################################
    # @mouse.when_clicked callbacks
    ####################################
    if (
        mouse_state.click_happened_this_frame
        and callback_manager.get_callbacks(CallbackType.WHEN_CLICKED) is not None
    ):
        for callback in callback_manager.get_callbacks(CallbackType.WHEN_CLICKED):
            await run_async_callback(
                callback,
                [],
                [],
            )

    ########################################
    # @mouse.when_click_released callbacks
    ########################################
    if (
        mouse_state.click_release_happened_this_frame
        and callback_manager.get_callbacks(CallbackType.WHEN_CLICK_RELEASED) is not None
    ):
        for callback in callback_manager.get_callbacks(
            CallbackType.WHEN_CLICK_RELEASED
        ):
            await run_async_callback(
                callback,
                [],
                [],
            )
